from __future__ import absolute_import, print_function
from unittest import TestCase, main
import io
import gzip
import six
import tempfile
from .io_helpers import GzipCompressStream, GzipDecompressStream, \
    CombineStreams, SplitStream, IteratorMixin
from six.moves.urllib.parse import urljoin
from six.moves.urllib.request import pathname2url, urlopen
from contextlib import closing


def path2url(path):
    return urljoin('file:', pathname2url(path))


try:
    from itertools import izip
except ImportError:
    izip = zip

mebibyte = 1024 * 1024
small_chunk = 10


class FakeMixinUser(IteratorMixin):
    def __init__(self, data, block_size):
        self.backing_buffer = io.BytesIO(data)
        self._IteratorMixin__block_size = block_size

    def read(self, n):
        return self.backing_buffer.read(n)

MIXIN_DATA = b"""a
aa
aaa
aaaa
aaaaa
aaaaaa
"""


class IOHelpersTest(TestCase):
    def test_gzip_compress_stream(self):
        underlying = io.BytesIO(b"This is a short test stream")
        instance = GzipCompressStream(underlying, block_size=4)
        result = b""
        for i in range(1000):
            read_result = instance.read(i)
            result += read_result
            if len(read_result) < i:
                break
        self.assertEqual(b"", instance.read(10))
        self.assertEqual(b"", instance.read())

        # Decode with GzipFile
        test_decoder = gzip.GzipFile('', mode='r', fileobj=io.BytesIO(result))
        self.assertEqual(b"This is a short test stream", test_decoder.read())

    def test_gzip_compress_stream_read_all(self):
        underlying = io.BytesIO(b"This is a short test stream")
        instance = GzipCompressStream(underlying, block_size=4)
        result = instance.read()
        self.assertEqual(b"", instance.read(10))
        self.assertEqual(b"", instance.read())

        # Decode with GzipFile
        test_decoder = gzip.GzipFile('', mode='r', fileobj=io.BytesIO(result))
        self.assertEqual(b"This is a short test stream", test_decoder.read())

    def test_gzip_decompress_stream(self):
        underlying = io.BytesIO()
        gzwriter = gzip.GzipFile(fileobj=underlying, mode='w')
        gzwriter.write(b"This is a short test stream")
        gzwriter.close()
        underlying.seek(0)

        instance = GzipDecompressStream(underlying, block_size=4)
        result = b""
        for i in range(1000):
            read_result = instance.read(i)
            result += read_result
            if len(read_result) < i:
                break
        self.assertEqual(b"", instance.read(10))
        self.assertEqual(b"", instance.read())

        self.assertEqual(b"This is a short test stream", result)

    def test_gzip_decompress_stream_read_all(self):
        underlying = io.BytesIO()
        gzwriter = gzip.GzipFile(fileobj=underlying, mode='w')
        gzwriter.write(b"This is a short test stream")
        gzwriter.close()
        underlying.seek(0)

        instance = GzipDecompressStream(underlying, block_size=4)
        result = instance.read()
        self.assertEqual(b"", instance.read(10))
        self.assertEqual(b"", instance.read())

        self.assertEqual(b"This is a short test stream", result)

    def test_split_file(self):
        with tempfile.TemporaryFile() as f:
            f.write(b"a" * mebibyte)
            f.write(b"b" * mebibyte)
            f.flush()
            f.seek(0)
            instance = SplitStream(f, mebibyte)
            self.assertEqual(len(instance), 2)
            for m, c in izip(instance, [b"a", b"b"]):
                self.assertEqual(m.read(mebibyte), c * mebibyte)

    def test_split_file_irregular_size(self):
        with tempfile.TemporaryFile() as f:
            f.write(b"a" * mebibyte)
            f.write(b"b")
            f.flush()
            f.seek(0)
            instance = SplitStream(f, mebibyte)
            self.assertEqual(len(instance), 2)
            iterator = iter(instance)
            chunk1 = six.next(iterator)
            self.assertEqual(b"a" * mebibyte, chunk1.read(mebibyte))
            chunk2 = six.next(iterator)
            self.assertEqual(b"b", chunk2.read(mebibyte))

    def test_split_file_irregular_size_2(self):
        with tempfile.TemporaryFile() as f:
            f.write(b"a" * mebibyte)
            f.write(b"b" * (mebibyte - 1))
            f.flush()
            f.seek(0)
            instance = SplitStream(f, mebibyte)
            self.assertEqual(len(instance), 2)
            iterator = iter(instance)
            chunk1 = six.next(iterator)
            self.assertEqual(b"a" * mebibyte, chunk1.read(mebibyte))
            chunk2 = six.next(iterator)
            self.assertEqual(b"b" * (mebibyte - 1), chunk2.read(mebibyte))

    def test_split_url(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b"a" * mebibyte)
            f.write(b"b" * mebibyte)
            f.flush()
            with closing(urlopen(path2url(f.name))) as stream:
                instance = SplitStream(stream, mebibyte)
                self.assertEqual(len(instance), 2)
                for m, c in izip(instance, [b"a", b"b"]):
                    self.assertEqual(m.read(mebibyte), c * mebibyte)

    def test_split_url_irregular_size(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b"a" * mebibyte)
            f.write(b"b")
            f.flush()
            with closing(urlopen(path2url(f.name))) as stream:
                instance = SplitStream(stream, mebibyte)
                self.assertEqual(len(instance), 2)
                iterator = iter(instance)
                chunk1 = six.next(iterator)
                self.assertEqual(b"a" * mebibyte, chunk1.read(mebibyte))
                chunk2 = six.next(iterator)
                self.assertEqual(b"b", chunk2.read(mebibyte))

    def test_split_url_irregular_size_2(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b"a" * mebibyte)
            f.write(b"b" * (mebibyte - 1))
            f.flush()
            with closing(urlopen(path2url(f.name))) as stream:
                instance = SplitStream(stream, mebibyte)
                self.assertEqual(len(instance), 2)
                iterator = iter(instance)
                chunk1 = six.next(iterator)
                self.assertEqual(b"a" * mebibyte, chunk1.read(mebibyte))
                chunk2 = six.next(iterator)
                self.assertEqual(b"b" * (mebibyte - 1), chunk2.read(mebibyte))

    def test_split_bytes(self):
        instance = SplitStream(b"a" * small_chunk + b"b" * small_chunk,
                               small_chunk)
        self.assertEqual(len(instance), 2)
        for m, c in izip(instance, [b"a", b"b"]):
            self.assertEqual(c * small_chunk, m.read(small_chunk))

    def test_split_bytes_irregular_size(self):
        instance = SplitStream(b"a" * small_chunk + b"b", small_chunk)
        self.assertEqual(len(instance), 2)
        iterator = iter(instance)
        chunk1 = six.next(iterator)
        self.assertEqual(b"a" * small_chunk, chunk1.read(small_chunk))
        chunk2 = six.next(iterator)
        self.assertEqual(b"b", chunk2.read(small_chunk))

    def test_split_bytes_irregular_size_2(self):
        instance = SplitStream(b"a" * small_chunk + b"b" * (small_chunk - 1),
                               small_chunk)
        self.assertEqual(len(instance), 2)
        iterator = iter(instance)
        chunk1 = six.next(iterator)
        self.assertEqual(b"a" * small_chunk, chunk1.read(small_chunk))
        chunk2 = six.next(iterator)
        self.assertEqual(b"b" * (small_chunk - 1), chunk2.read(small_chunk))

    def test_split_combine_stream_misaligned_with_chunk_size_1(self):
        instance = SplitStream(dict(
            Body=CombineStreams([io.BytesIO(b"1234"), io.BytesIO(b"567890123456789")]),
            ContentLength=19
        ), 5)
        self.assertEqual(len(instance), 4)
        iterator = iter(instance)
        chunk1 = six.next(iterator)
        self.assertEqual(b"12345", chunk1.read(5))
        chunk2 = six.next(iterator)
        self.assertEqual(b"67890", chunk2.read(5))
        chunk3 = six.next(iterator)
        self.assertEqual(b"12345", chunk3.read(5))
        chunk4 = six.next(iterator)
        self.assertEqual(b"6789", chunk4.read(5))

    def test_split_combine_stream_misaligned_with_chunk_size_2(self):
        instance = SplitStream(dict(
            Body=CombineStreams([io.BytesIO(b"123456789"), io.BytesIO(b"012345678")]),
            ContentLength=18
        ), 5)
        self.assertEqual(len(instance), 4)
        iterator = iter(instance)
        chunk1 = six.next(iterator)
        self.assertEqual(b"12345", chunk1.read(5))
        chunk2 = six.next(iterator)
        self.assertEqual(b"67890", chunk2.read(5))
        chunk3 = six.next(iterator)
        self.assertEqual(b"12345", chunk3.read(5))
        chunk4 = six.next(iterator)
        self.assertEqual(b"678", chunk4.read(5))

    def test_split_combine_stream_misaligned_with_chunk_size_3(self):
        instance = SplitStream(dict(
            Body=CombineStreams([io.BytesIO(b"123"), io.BytesIO(b"456"), io.BytesIO(b"789"), io.BytesIO(b"012")]),
            ContentLength=12
        ), 5)
        self.assertEqual(len(instance), 3)
        iterator = iter(instance)
        chunk1 = six.next(iterator)
        self.assertEqual(b"12345", chunk1.read(5))
        chunk2 = six.next(iterator)
        self.assertEqual(b"67890", chunk2.read(5))
        chunk3 = six.next(iterator)
        self.assertEqual(b"12", chunk3.read(5))

    def test_combine_streams(self):
        instance = CombineStreams(io.BytesIO(b"Hello") for i in range(20))
        result = b""
        for i in range(1000):
            read_result = instance.read(i)
            result += read_result
            if len(read_result) < i:
                break

        self.assertEqual(instance.read(10), b"")
        self.assertEqual(instance.read(), b"")
        self.assertEqual(result, b"Hello" * 20)

    def test_combine_streams_readall(self):
        instance = CombineStreams(io.BytesIO(b"Hello") for i in range(20))
        result = instance.read()

        self.assertEqual(instance.read(10), b"")
        self.assertEqual(instance.read(), b"")
        self.assertEqual(result, b"Hello" * 20)

    def test_iterator_mixin(self):
        # import pudb
        # pu.db
        instance = FakeMixinUser(MIXIN_DATA, 4)
        self.assertEqual(b'a\n', instance.readline())
        self.assertEqual(b'aa\n', instance.readline())
        self.assertEqual(b'aaa\n', instance.readline())
        self.assertEqual(b'aaaa\n', instance.readline())
        self.assertEqual(b'aaaaa\n', instance.readline())
        self.assertEqual(b'aaaaaa\n', instance.readline())
        self.assertEqual(b'', instance.readline())

    def test_iterator_mixin_list(self):
        # import pudb
        # pu.db
        instance = FakeMixinUser(MIXIN_DATA, 4)
        self.assertEqual([
            b'a\n',
            b'aa\n',
            b'aaa\n',
            b'aaaa\n',
            b'aaaaa\n',
            b'aaaaaa\n',
        ], instance.readlines())


if __name__ == '__main__':
    main()



from unittest import TestCase, main, skip
import mesh_client
from requests import HTTPError


class EndpointConnectivityTest(TestCase):
    @skip("This test needs N3 connectivity, talks to real endpoints, and is Python 3 only, so don't run it by default")
    def test_connectivity(self):
        for key, endpoint in list(mesh_client.__dict__.items()):
            if key.endswith('_ENDPOINT') and not key.startswith('LOCAL_') and not '_OPENTEST_' in key:
                with self.subTest(key):
                    print("Testing", key)
                    with self.assertRaises(HTTPError) as cm:
                        client = mesh_client.MeshClient(endpoint, 'BADUSERNAME', 'BADPASSWORD')
                        client.handshake()
                    self.assertEqual(cm.exception.response.status_code, 400)


if __name__ == '__main__':
    main()



#!/usr/bin/python3
"""
A crude mock implementation of the MESH API, suitable for use in
small-scale tests. This file can either be run as a script, which will serve
on port 8000, or used as a test fixture, which will serve on a random port.

To run mock mesh as a script, run:

python -m mesh_client.mock_server

This mock instance will use TLS mutual authentication, with hard-coded SSL
certs that are included in the distribution. Matching client certs are
also included, and the settings to use them are included in the mesh_client
package as default_ssl_opts. Since these certs and keys are publicly available,
they should only be used in test environments.
"""
from __future__ import print_function, absolute_import
import datetime
import hmac
import json
import os.path
import random
import ssl
import time
import traceback
import zlib
from contextlib import closing
from hashlib import sha256
from collections import OrderedDict
from threading import Thread
from wsgiref.util import shift_path_info
from wsgiref.simple_server import make_server, WSGIServer
from .io_helpers import stream_from_wsgi_environ
from .key_helper import get_shared_key_from_environ

_OPTIONAL_HEADERS = {
    "HTTP_CONTENT_ENCODING": "Content-Encoding",
    "HTTP_MEX_WORKFLOWID": "Mex-WorkflowID",
    "HTTP_MEX_FILENAME": "Mex-FileName",
    "HTTP_MEX_LOCALID": "Mex-LocalID",
    "HTTP_MEX_MESSAGETYPE": "Mex-MessageType",
    "HTTP_MEX_PROCESSID": "Mex-ProcessID",
    "HTTP_MEX_PROCESSLID": "Mex-ProcesslID",
    "HTTP_MEX_SUBJECT": "Mex-Subject",
    "HTTP_MEX_CONTENT_ENCRYPTED": "Mex-Content-Encrypted",
    "HTTP_MEX_CONTENT_COMPRESS": "Mex-Content-Compress",
    "HTTP_MEX_CONTENT_CHECKSUM": "Mex-Content-Checksum",
    "HTTP_MEX_COMPRESSED": "Mex-Compressed",
    "HTTP_MEX_CONTENT_COMPRESSED": "Mex-Content-Compressed",
    "HTTP_MEX_CHUNK_RANGE": "Mex-Chunk-Range",
    "HTTP_MEX_FROM": "Mex-From",
    "HTTP_MEX_TO": "Mex-To"
}


TIMESTAMP_FORMAT = '%Y%m%d%H%M%S%f'


def _dumb_response_code(code, message):
    def handle(environ, start_response):
        start_response("{} {}".format(code, message),
                       [("Content-Type", "text/plain")])
        return [message.encode("UTF-8")]

    return handle


_not_found = _dumb_response_code(404, "Not Found")
_not_authorized = _dumb_response_code(401, "Unauthorized")
_forbidden = _dumb_response_code(403, "Forbidden")
_simple_ok = _dumb_response_code(200, "OK")
_bad_request = _dumb_response_code(400, "Bad Request")
_server_error = _dumb_response_code(500, "Server Error")


def _ok(content_type, data, start_response):
    start_response("200 OK", [("Content-Type", content_type)])
    return data


def _error(content_type, data, start_error_response):
    start_error_response("500 INTERNAL SERVER ERROR", [("Content-Type", content_type)])
    return data


def _compose(**kwargs):
    def handle(environ, start_response):
        path_component = shift_path_info(environ) or 'default'
        if path_component in kwargs:
            return kwargs[path_component](environ, start_response)
        else:
            return _not_found(environ, start_response)

    return handle


class MockMeshApplication:
    """
    A crude mock instance of MESH, suitable for use in tests.

    This class can be used as a context manager:

    with MockMeshApplication() as app:
        uri = app.uri
        # Send and receive stuff from MESH endpoint at uri
        assert app.messages == expectedMessages

    This will spin up a MESH instance on a random port, and make its uri
    available as app.uri. The instance will be shut down when the context
    manager exits. These capabilities are also available as start_server and
    stop_server, for use in cases where context managers are not convenient.

    Subclasses may wish to override make_message_id, to allow testing with
    different message id formats.
    """

    def __init__(self, shared_key=get_shared_key_from_environ()):
        self.messages = {}
        self._shared_key = shared_key
        self._msg_count = 0

    @property
    def __call__(self):
        return _compose(
            messageexchange=self.message_exchange,
            endpointlookup=_compose(
                mesh=self.endpoint_lookup
            )
        )

    def authenticated(self, handler):
        def handle(environ, start_response):
            requested_mailbox = shift_path_info(environ)
            authorization_header = environ.get("HTTP_AUTHORIZATION", "")
            if not authorization_header.startswith("NHSMESH "):
                return _not_authorized(environ, start_response)

            auth_data = authorization_header[8:]
            mailbox, nonce, nonce_count, ts, hashed = auth_data.split(":")
            expected_password = "password"
            hash_data = ":".join([
                mailbox, nonce, nonce_count, expected_password, ts
            ])
            myhash = hmac.HMAC(self._shared_key, hash_data.encode("ASCII"),
                               sha256).hexdigest()
            if myhash == hashed and mailbox == requested_mailbox:
                environ["mesh.mailbox"] = mailbox
                return handler(environ, start_response)
            else:
                return _forbidden(environ, start_response)

        return handle

    def handshake(self, environ, start_response):
        auth_headers = [
            'HTTP_MEX_CLIENTVERSION',
            'HTTP_MEX_JAVAVERSION',
            'HTTP_MEX_OSARCHITECTURE',
            'HTTP_MEX_OSNAME',
            'HTTP_MEX_OSVERSION'
        ]
        for auth_header in auth_headers:
            if auth_header not in environ:
                return _server_error(environ, start_response)
        mailbox_id = environ["mesh.mailbox"]
        return _ok('application/json',
                   [json.dumps({"mailboxId": mailbox_id}).encode("UTF-8")],
                   start_response)

    @property
    def message_exchange(self):
        return self.authenticated(
            _compose(
                inbox=self.inbox,
                outbox=self.outbox,
                count=self.count,
                default=self.handshake
            )
        )


    def inbox(self, environ, start_response):
        request_method = environ["REQUEST_METHOD"]
        message_id = shift_path_info(environ)
        mailbox = environ["mesh.mailbox"]

        if (request_method == "PUT" and
                environ["PATH_INFO"] == "/status/acknowledged"):
            del self.messages[mailbox][message_id]
            return _simple_ok(environ, start_response)

        if request_method == "GET":
            if message_id:
                chunk_num = shift_path_info(environ)
                if chunk_num:
                    return self.download_chunk(
                        message_id, chunk_num)(environ, start_response)
                message = self.messages[message_id]
                if message['headers']['Mex-To'] != mailbox:
                    return _not_found(environ, start_response)
                response_code = ("206 Partial Content"
                                 if "chunks" in message else "200 OK")
                start_response(response_code, list(message["headers"].items()))
                return [message["data"]]
            else:
                messages = {"messages": list(
                    self.messages.get(mailbox, {}).keys())}
                return _ok("application/json",
                           [json.dumps(messages).encode("UTF-8")],
                           start_response)
        else:
            return _bad_request(environ, start_response)

    def count(self, environ, start_response):
        request_method = environ["REQUEST_METHOD"]
        mailbox = environ["mesh.mailbox"]
        if request_method == "GET":
            result = {
                "count": len(self.messages.get(mailbox, {})),
                "internalID": "12345",
                "allResultsIncluded": True,

            }
            return _ok("application/json",
                       [json.dumps(result).encode("UTF-8")],
                       start_response)
        else:
            return _bad_request(environ, start_response)

    def outbox(self, environ, start_response):
        chunk_msg = shift_path_info(environ)
        if chunk_msg == 'tracking':
            return self.tracking(environ, start_response)
        if chunk_msg:
            return self.upload_chunk(chunk_msg)(environ, start_response)
        try:
            recipient = environ["HTTP_MEX_TO"]
            sender = environ["HTTP_MEX_FROM"]
            mailbox_id = environ["mesh.mailbox"]
            assert mailbox_id == sender
        except Exception as e:
            traceback.print_exc()
            start_response("417 Expectation Failed",
                           [('Content-Type', 'application/json')])
            return [json.dumps({
                "errorCode": "02",
                "errorDescription": str(e),
                "errorEvent": "COLLECT",
                "messageID": "99999"
            })]

        mailbox = self.messages.setdefault(recipient, OrderedDict())
        with closing(stream_from_wsgi_environ(environ)) as stream:
            data = stream.read()
        if not data:
            start_response("417 Expectation Failed",
                           [('Content-Type', 'application/json')])
            return [json.dumps({
                "errorCode": "02",
                "errorDescription": "Data file is missing or inaccessible.",
                "errorEvent": "COLLECT",
                "messageID": "99999"
            }).encode("utf-8")]
        headers = {_OPTIONAL_HEADERS[key]: value
                   for key, value in environ.items()
                   if key in _OPTIONAL_HEADERS}
        msg_id = self.make_message_id()
        headers['Mex-MessageID'] = msg_id
        mailbox[msg_id] = {"headers": headers, "data": data}
        self.messages[msg_id] = mailbox[msg_id]
        return _ok("application/json",
                   [json.dumps({"messageID": msg_id}).encode("UTF-8")],
                   start_response)

    def upload_chunk(self, chunk_msg):
        def handle(environ, start_response):
            chunk_num = shift_path_info(environ)
            msg = self.messages[chunk_msg]
            chunks = msg.setdefault("chunks", {})
            with closing(stream_from_wsgi_environ(environ)) as stream:
                data = stream.read()
            chunks[chunk_num] = zlib.compress(data) if not environ.get('HTTP_CONTENT_ENCODING', None) else data
            start_response('202 Accepted', [])
            return []

        return handle

    def download_chunk(self, chunk_msg, chunk_num):
        def handle(environ, start_response):
            msg = self.messages[chunk_msg]
            chunks = msg["chunks"]
            chunk = chunks[chunk_num]
            if environ['HTTP_ACCEPT_ENCODING'] != "gzip":
                chunk = zlib.decompress(chunk)

            chunk_header = "{}:{}".format(chunk_num, len(chunks) + 1)
            start_response('200 OK', [
                ('Content-Type', 'application/octet-stream'),
                ('Content-Encoding', 'gzip'),
                ('Mex-Chunk-Range', chunk_header)
            ])
            return [chunk]

        return handle

    def tracking(self, environ, start_response):
        tracking_id = shift_path_info(environ)
        if not tracking_id:
            return _bad_request(environ, start_response)
        for message in self.messages.values():
            if message.get('headers', {}).get('Mex-LocalID') == tracking_id:
                headers = message['headers']
                recipient = headers['Mex-To']
                message_id = headers['Mex-MessageID']
                acked = message_id not in self.messages[recipient]
                now = datetime.datetime.now().strftime(TIMESTAMP_FORMAT)
                response = {
                    "localId": tracking_id,
                    "dtsId": message_id,
                    "messageType": "DATA",
                    "chunkCount": len(message['chunks']),
                    "subject": headers.get('Mex-Subject'),
                    "statusEvent": None,
                    "version": "1.0",
                    "downloadTimestamp": now,
                    "statusDescription": None,
                    "status": "Acknowledged" if acked else "Accepted",
                    "workflowId": headers.get('Mex-WorkflowID'),
                    "fileName": headers.get('Mex-FileName'),
                    "uploadTimestamp": now,
                    "recipient": headers['Mex-To'],
                    "sender": headers['Mex-From'],
                    "messageId": message_id,
                    "fileSize": 0
                }
                return _ok('application/json',
                           [json.dumps(response).encode('UTF-8')],
                           start_response)
        else:
            return _not_found(environ, start_response)

    def endpoint_lookup(self, environ, start_response):
        org_code = shift_path_info(environ)
        workflow_id = shift_path_info(environ)
        if not org_code or not workflow_id:
            return _not_found(environ, start_response)
        result = {
            "query_id": "{ts:%Y%m%d%H%M%S%f}_{rnd:06x}_{ts:%s}".format(
                ts=datetime.datetime.now(),
                rnd=random.randint(0, 0xffffff)),
            "results": [
                {
                    "address": "{}HC001".format(org_code),
                    "description": "{} {} endpoint".format(org_code, workflow_id),
                    "endpoint_type": "MESH"
                }
            ]
        }
        return _ok('application/json',
                   [json.dumps(result).encode('UTF-8')],
                   start_response)

    def make_message_id(self):
        """
        Create a new message id. By default, this uses roughly the same format
        as MESH (although the format is undocumented), but it can be overridden
        in subclasses for different behaviour
        """
        n = self._msg_count
        self._msg_count += 1
        return "{ts:%Y%m%d%H%M%S%f}_{num:06d}".format(
            ts=datetime.datetime.now(),
            num=n
        )

    def __enter__(self):
        self.start_server()
        return self

    def __exit__(self, ex_type, value, tb):
        self.stop_server()

    def start_server(self):
        """
        Start an embedded web server, running the mock MESH app. After running
        this method, the uri of the running MESH instance will be available
        as self.uri
        """
        self.server = make_server("", 0, self, server_class=SSLWSGIServer)
        port = self.server.server_address[1]
        self.uri = "https://localhost:{}".format(port)
        thread = Thread(
            target=self.server.serve_forever,
            kwargs={"poll_interval": 0.01})
        thread.daemon = True
        thread.start()
        return self

    def stop_server(self):
        """
        Stop a server that was started by start_server(). This has no effect
        when the app is run from an external WSGI container, such as uWSGI,
        GUnicorn or Cheroot.
        """
        self.server.shutdown()


class MockMeshChunkRetryApplication(MockMeshApplication, object):
    def __init__(self):
        self.allowed_retries = {}
        super(MockMeshChunkRetryApplication, self).__init__()

    def set_chunk_retry_options(self, options):
        self.retry_options = options

    def outbox(self, environ, start_response):
        current_chunk = int(environ['HTTP_MEX_CHUNK_RANGE'].split(':')[0])
        if current_chunk == 1:
            self.allowed_retries = {p[0]: p[1] for p in self.retry_options}

        if current_chunk in self.allowed_retries:
            if self.allowed_retries[current_chunk] >= 0:
                self.allowed_retries[current_chunk] -= 1
                return _error("application/json", '', start_response)

        return super(MockMeshChunkRetryApplication, self).outbox(environ, start_response)


class SlowMockMeshApplication(MockMeshApplication):
    def __call__(self, environ, start_response):
        time.sleep(2.0)
        return super(SlowMockMeshApplication, self).__call__(environ, start_response)


_data_dir = os.path.dirname(__file__)
default_server_context = ssl.create_default_context(
    ssl.Purpose.CLIENT_AUTH, cafile=os.path.join(_data_dir, "ca.cert.pem"))
default_server_context.load_cert_chain(
    os.path.join(_data_dir, 'server.cert.pem'),
    os.path.join(_data_dir, 'server.key.pem'))
default_server_context.check_hostname = False
default_server_context.verify_mode = ssl.CERT_REQUIRED


class SSLWSGIServer(WSGIServer, object):
    __context = default_server_context

    def get_request(self):
        (socket, addr) = super(SSLWSGIServer, self).get_request()
        return self.__context.wrap_socket(socket, server_side=True), addr


def main():
    print("Serving on port 8000")
    server = make_server(
        "", 8000, MockMeshApplication(), server_class=SSLWSGIServer)
    server.serve_forever(0.01)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
from __future__ import print_function
from mesh_client import AuthTokenGenerator
from mesh_client.key_helper import get_shared_key_from_environ
from argparse import ArgumentParser


def main():
    parser = ArgumentParser(description="Create MESH auth token for user")
    parser.add_argument('user', help="The username")
    parser.add_argument('password', help="The password for the user")
    parser.add_argument('--shared-key',
                        help="The shared key to use for MESH authentication",
                        default=get_shared_key_from_environ())
    args = parser.parse_args()
    generator = AuthTokenGenerator(args.shared_key, args.user, args.password)
    print(generator.generate_token())


if __name__ == '__main__':
    main()



from __future__ import absolute_import
import codecs
import collections
import uuid
import hmac
import datetime
import os.path
import pkg_resources
import platform
import requests
import six
import time
import warnings
from io import BytesIO
from itertools import chain
from hashlib import sha256
from .io_helpers import \
    CombineStreams, SplitStream, GzipCompressStream, GzipDecompressStream
from .key_helper import get_shared_key_from_environ

MOCK_CA_CERT = pkg_resources.resource_filename('mesh_client', "ca.cert.pem")
MOCK_CERT = pkg_resources.resource_filename('mesh_client', "client.cert.pem")
MOCK_KEY = pkg_resources.resource_filename('mesh_client', "client.key.pem")

MOCK_SSL_OPTS = {
    "verify": MOCK_CA_CERT,
    "cert": (MOCK_CERT, MOCK_KEY)
}
"""
Usable default values for verify and cert, providing certificates and keys
which should work with mock_server. Note that these certs will not work with
any NHS Digital test environments - such certs must be obtained from
NHS Digital.
"""
default_ssl_opts = MOCK_SSL_OPTS

INT_CA_CERT = pkg_resources.resource_filename('mesh_client', "nhs-int-ca-bundle.pem")
DEV_CA_CERT = pkg_resources.resource_filename('mesh_client', "nhs-dev-ca-bundle.pem")
DEP_CA_CERT = pkg_resources.resource_filename('mesh_client', "nhs-dep-ca-bundle.pem")
TRAIN_CA_CERT = pkg_resources.resource_filename('mesh_client', "nhs-train-ca-bundle.pem")
LIVE_CA_CERT = pkg_resources.resource_filename('mesh_client', "nhs-live-root-ca.pem")
OPENTEST_CA_CERT = pkg_resources.resource_filename('mesh_client', "nhs-opt-ca-bundle.pem")
DIGICERT_CA_CERT = pkg_resources.resource_filename('mesh_client', "nhs-digicert-ca-bundle.pem")
IG_INT_CA_CERT = pkg_resources.resource_filename('mesh_client', "nhs-ig-int-ca-bundle.pem")


_OPTIONAL_HEADERS = {
    "workflow_id": "Mex-WorkflowID",
    "filename": "Mex-FileName",
    "local_id": "Mex-LocalID",
    "message_type": "Mex-MessageType",
    "process_id": "Mex-ProcessID",
    "subject": "Mex-Subject",
    "encrypted": "Mex-Content-Encrypted",
    "compressed": "Mex-Content-Compressed",
    "checksum": "Mex-Content-Checksum"
}

_RECEIVE_HEADERS = {
    "sender": "Mex-From",
    "recipient": "Mex-To",
    "message_id": "Mex-MessageID",
    "version": "Mex-Version",
    "partner_id": "Mex-PartnerID",
    "recipient_smtp": "Mex-ToSMTP",
    "sender_smtp": "Mex-FromSMTP",
}
_RECEIVE_HEADERS.update(_OPTIONAL_HEADERS)


VERSION = pkg_resources.get_distribution('mesh_client').version


Endpoint = collections.namedtuple('Endpoint', ['url', 'verify', 'cert'])
LOCAL_MOCK_ENDPOINT = Endpoint('https://localhost:8000', MOCK_CA_CERT, (MOCK_CERT, MOCK_KEY))
LOCAL_FAKE_ENDPOINT = Endpoint('https://localhost:8829', MOCK_CA_CERT, (MOCK_CERT, MOCK_KEY))
NHS_INT_ENDPOINT = Endpoint('https://msg.int.spine2.ncrs.nhs.uk', INT_CA_CERT, None)
NHS_DEV_ENDPOINT = Endpoint('https://msg.dev.spine2.ncrs.nhs.uk', DEV_CA_CERT, None)
NHS_DEP_ENDPOINT = Endpoint('https://msg.dep.spine2.ncrs.nhs.uk', DEP_CA_CERT, None)
NHS_TRAIN_ENDPOINT = Endpoint('https://msg.train.spine2.ncrs.nhs.uk', TRAIN_CA_CERT, None)
NHS_LIVE_ENDPOINT = Endpoint('https://mesh-sync.national.ncrs.nhs.uk', LIVE_CA_CERT, None)
NHS_OPENTEST_ENDPOINT = Endpoint('https://192.168.128.11', OPENTEST_CA_CERT, None)
NHS_INTERNET_GATEWAY_ENDPOINT = Endpoint('https://mesh.spineservices.nhs.uk', DIGICERT_CA_CERT, None)
NHS_INTERNET_GATEWAY_INT_ENDPOINT = Endpoint('https://msg.intspineservices.nhs.uk', IG_INT_CA_CERT, None)


class MeshError(Exception):
    pass


class MeshClient(object):
    """
    A class representing a single MESH session, for a given user on a given
    endpoint. This class handles details such as chunking and compression
    transparently.
    """

    def __init__(self,
                 url,
                 mailbox,
                 password,
                 shared_key=get_shared_key_from_environ(),
                 cert=None,
                 verify=None,
                 max_chunk_size=75 * 1024 * 1024,
                 proxies=None,
                 transparent_compress=False,
                 max_chunk_retries=0,
                 timeout=10*60):
        """
        Create a new MeshClient.

        At a minimum, you must provide an endpoint url, a mailbox and a
        password. The endpoint URL can either be a string, or a preconfigured
        endpoint. Currently the following endpoints are preconfigured:

        LOCAL_MOCK_ENDPOINT
        LOCAL_FAKE_ENDPOINT
        NHS_INT_ENDPOINT
        NHS_LIVE_ENDPOINT
        NHS_OPENTEST_ENDPOINT
        NHS_INTERNET_ENDPOINT

        Since MESH uses mutual authentication, it is also highly
        advisable to provide SSL information, in the form of cert and verify.
        these take the same format as in the requests library, so you would
        typically provide a filename for a CA cert as verify, and a tuple
        containing two filenames (a client cert and a private key) for cert.

        If you have chosen to use a preconfigured endpoint, then you a sane
        default value will be used for the CA cert, so you should not have to
        configure verify. For mock and fake endpoints, default values for cert
        are provided, so you will not need to configure that either.

        You can also optionally specify the maximum file size before chunking,
        and whether messages should be compressed, transparently, before
        sending.
        """
        self._session = requests.Session()
        self._session.headers = {
            "mex-ClientVersion": "mesh_client=={}".format(VERSION),
            "mex-OSArchitecture": platform.processor(),
            "mex-OSName": platform.system(),
            "mex-OSVersion": "{} {}".format(platform.release(), platform.version()),
            "mex-JavaVersion": "N/A",
            "Accept-Encoding": "gzip"
        }
        self._session.auth = AuthTokenGenerator(shared_key, mailbox, password)
        if hasattr(url, 'url'):
            self._url = url.url
        else:
            self._url = url

        if hasattr(url, 'cert') and cert is None:
            self._session.cert = url.cert
        else:
            self._session.cert = cert

        if hasattr(url, 'verify') and verify is None:
            self._session.verify = url.verify
        else:
            self._session.verify = verify

        self._session.proxies = proxies or {}
        self._mailbox = mailbox
        self._max_chunk_size = max_chunk_size
        self._transparent_compress = transparent_compress
        self._max_chunk_retries = max_chunk_retries
        self._timeout = timeout
        self._close_called = False

    def handshake(self):
        """
        List all messages in user's inbox. Returns a list of message_ids
        """
        response = self._session.post(
            "{}/messageexchange/{}".format(self._url, self._mailbox),
            timeout=self._timeout
        )

        response.raise_for_status()

        return b'hello'

    def count_messages(self):
        """
        Count all messages in user's inbox. Returns an integer
        """
        response = self._session.get(
            "{}/messageexchange/{}/count".format(self._url, self._mailbox),
            timeout=self._timeout)
        response.raise_for_status()
        return response.json()["count"]

    def get_tracking_info(self, tracking_id):
        """
        Gets tracking information from MESH about a message, by its local message id.
        Returns a dictionary, in much the same format that MESH provides it.
        """
        response = self._session.get(
            "{}/messageexchange/{}/outbox/tracking/{}".format(self._url, self._mailbox, tracking_id),
            timeout=self._timeout)
        response.raise_for_status()
        return response.json()

    def lookup_endpoint(self, organisation_code, workflow_id):
        """
        Lookup a mailbox by organisation code and workflow id.
        Returns a dictionary, in much the same format that MESH provides it.
        """
        response = self._session.get(
            "{}/endpointlookup/mesh/{}/{}".format(self._url, organisation_code, workflow_id),
            timeout=self._timeout)
        response.raise_for_status()
        return response.json()

    def list_messages(self):
        """
        List all messages in user's inbox. Returns a list of message_ids
        """
        response = self._session.get(
            "{}/messageexchange/{}/inbox".format(self._url, self._mailbox),
            timeout=self._timeout)
        response.raise_for_status()
        return response.json()["messages"]

    def retrieve_message(self, message_id):
        """
        Retrieve a message based on its message_id. This will return a Message
        object
        """
        message_id = getattr(message_id, "_msg_id", message_id)
        response = self._session.get(
            "{}/messageexchange/{}/inbox/{}".format(self._url, self._mailbox, message_id),
            stream=True,
            timeout=self._timeout)
        response.raise_for_status()
        return Message(message_id, response, self)

    def retrieve_message_chunk(self, message_id, chunk_num):
        response = self._session.get(
            "{}/messageexchange/{}/inbox/{}/{}".format(self._url, self._mailbox, message_id, chunk_num),
            stream=True,
            timeout=self._timeout)
        response.raise_for_status()
        return response

    def send_message(self,
                     recipient,
                     data,
                     **kwargs):
        """
        Send a message to recipient containing data.

        This method optionally allows the user to provide the following keyword
        arguments, which specify properties of the message, and map to
        the equivalent properties in MESH - either headers or control file
        entries, depending on the type of consumer:

        workflow_id
        filename
        local_id
        message_type
        process_id
        subject
        encrypted
        compressed
        checksum
        sender
        recipient
        message_id
        version
        partner_id
        recipient_smtp
        sender_smtp

        Note that compressed refers to *non-transparent* compression - the
        client will not attempt to compress or decompress data. Transparent
        compression for sending is enabled as a constructor option.
        """
        transparent_compress = self._transparent_compress
        maybe_compressed = (
            lambda stream: GzipCompressStream(
                stream) if transparent_compress else stream
        )
        headers = {
            "Mex-From": self._mailbox,
            "Mex-To": recipient,
            "Mex-MessageType": 'DATA',
            "Mex-Version": '1.0'
        }

        for key, value in kwargs.items():
            if key in _OPTIONAL_HEADERS:
                headers[_OPTIONAL_HEADERS[key]] = str(value)
            else:
                raise TypeError("Unrecognised keyword argument {key}."
                                " optional arguments are: {args}".format(
                                    key=key,
                                    args=", ".join([
                                        "recipient", "data"
                                    ] + list(_OPTIONAL_HEADERS.keys()))))

        if transparent_compress:
            headers["Mex-Content-Compress"] = "TRUE"
            headers["Content-Encoding"] = "gzip"

        chunks = SplitStream(data, self._max_chunk_size)
        headers["Mex-Chunk-Range"] = "1:{}".format(len(chunks))
        chunk_iterator = iter(chunks)

        chunk1 = maybe_compressed(six.next(chunk_iterator))
        response1 = self._session.post(
            "{}/messageexchange/{}/outbox".format(self._url, self._mailbox),
            data=chunk1,
            headers=headers,
            timeout=self._timeout)
        json_resp = response1.json()
        if response1.status_code == 417 or "errorDescription" in json_resp:
            raise MeshError(json_resp["errorDescription"], json_resp)
        message_id = json_resp["messageID"]

        for i, chunk in enumerate(chunk_iterator):
            data = maybe_compressed(chunk)

            if self._max_chunk_retries > 0:
                if hasattr(data, 'read'):
                    data = data.read()
                buf = BytesIO(data)
            else:
                buf = data

            chunk_num = i + 2
            headers = {
                "Content-Type": "application/octet-stream",
                "Mex-Chunk-Range": "{}:{}".format(chunk_num, len(chunks)),
                "Mex-From": self._mailbox,
            }
            if transparent_compress:
                headers["Mex-Content-Compress"] = "TRUE"
                headers["Content-Encoding"] = "gzip"

            response = None
            for i in range(self._max_chunk_retries + 1):
                if self._max_chunk_retries > 0:
                    buf.seek(0)

                # non-linear delay in terms of squares
                time.sleep(i**2)

                response = self._session.post(
                    "{}/messageexchange/{}/outbox/{}/{}".format(
                        self._url, self._mailbox, message_id, chunk_num),
                    data=buf,
                    headers=headers,
                    timeout=self._timeout)

                # check other successful response codes
                if response.status_code == 200 or response.status_code == 202:
                    break
            else:
                response.raise_for_status()

        return message_id

    def acknowledge_message(self, message_id):
        """
        Acknowledge a message_id, deleting it from MESH.
        """
        message_id = getattr(message_id, "_msg_id", message_id)
        response = self._session.put(
            "{}/messageexchange/{}/inbox/{}/status/acknowledged".format(
                self._url, self._mailbox, message_id),
            timeout=self._timeout)
        response.raise_for_status()

    def iterate_all_messages(self):
        """
        Iterate over a list of Message objects for all messages in the user's
        inbox. This is provided as a convenience function, but will be
        slower than list_messages if only the message_ids are needed, since it
        will also begin to download messages.
        """
        for msg_id in self.list_messages():
            yield self.retrieve_message(msg_id)

    def close(self):
        self._close_called = True
        self._session.close()

    def __del__(self):
        if not self._close_called:
            warnings.warn(
                "The API of MeshClient changed in mesh_client 1.0. Each"
                " MeshClient instance must now be closed when the instance is"
                " no longer needed. This can be achieved by using the close"
                " method, or by using MeshClient in a with block. The"
                " connection pool will be closed for you by the destructor"
                " on this occasion, but you should not rely on this."
            )
            self.close()

    def __enter__(self):
        return self

    def __exit__(self, type_, value, tb):
        self.close()


class Message(object):
    """
    An object representing a message received from MESH. This is a file-like
    object, and can be passed to anything that expects an object with a `read`
    method.

    Any properties set on the message (as headers in MESH API, or control file
    entries) are available as attributes of this object. The following are
    supported:

    workflow_id
    filename
    local_id
    message_type
    process_id
    subject
    encrypted
    compressed

    Note that compressed refers to *non-transparent* compression - the
    client will not attempt to compress or decompress data. Transparent
    compression is handled automatically, with no intervention needed.

    Messages have a read method, and will handle chunking and transparent
    compression automatically. Once the data has been read, you must close the
    underlying stream using the close method. Data can only be read once. If
    you need to read it again, retrieve the message again.

    Messages can be used as context managers. When used in this way, streams
    will be closed automatically, and messages will be acknowledged if
    no exceptions are thrown whilst handling the message.
    """

    def __init__(self, msg_id, response, client):
        self._msg_id = msg_id
        self._client = client
        self._mex_headers = {}

        headers = response.headers
        for key, value in six.iteritems(headers):
            lkey = key.lower()
            if lkey.startswith('mex-'):
                self._mex_headers[lkey[4:]] = value

        for key, value in _RECEIVE_HEADERS.items():
            header_value = headers.get(value, None)
            if key in ["compressed", "encrypted"]:
                header_value = header_value or "FALSE"
                header_value = header_value.upper() == "TRUE"
            setattr(self, key, header_value)
        chunk, chunk_count = map(
            int, headers.get("Mex-Chunk-Range", "1:1").split(":"))
        maybe_decompress = (
            lambda resp:
            GzipDecompressStream(resp.raw)
            if resp.headers.get("Content-Encoding") == "gzip" else resp.raw
        )
        self._response = CombineStreams(
            chain([maybe_decompress(response)], (maybe_decompress(
                client.retrieve_message_chunk(msg_id, str(
                    i + 2))) for i in range(chunk_count - 1))))

    def id(self):
        """return the message id

        Returns:
            str: message id
        """
        return self._msg_id

    def read(self, n=None):
        """
        Read up to n bytes from the message, or read the remainder of the
        message, if n is not provided.
        """
        return self._response.read(n)

    def readline(self):
        """
        Read a single line from the message
        """
        return self._response.readline()

    def readlines(self):
        """
        Read all lines from the message
        """
        return self._response.readlines()

    def close(self):
        """Close the stream underlying this message"""
        if hasattr(self._response, "close"):
            try:
                self._response.close()
            finally:
                self._response = None

    def acknowledge(self):
        """
        Acknowledge this message, and delete it from MESH
        """
        self._client.acknowledge_message(self._msg_id)

    def mex_header(self, key, default=None):
        """ get a mex header if present

        Args:
            key (str): key
            default (any): default value
        Returns:
            str: the mex header value
        """
        return self._mex_headers.get(key, default)

    def mex_headers(self):
        """returns a generator iteritems for all the headers"""
        return six.iteritems(self._mex_headers)

    def __enter__(self):
        return self

    def __exit__(self, typ, value, traceback):
        try:
            if not value:
                self.acknowledge()
        finally:
            self.close()

    def __iter__(self):
        """
        Iterate through lines of the message
        """
        return iter(self._response)


class AuthTokenGenerator(object):

    def __init__(self, key, mailbox, password):
        self._key = key
        self._mailbox = mailbox
        self._password = password
        self._nonce = uuid.uuid4()
        self._nonce_count = 0

    def __call__(self, r=None):
        token = self.generate_token()
        if r is not None:
            # This is being used as a Requests auth handler
            r.headers['Authorization'] = token
            return r
        else:
            # This is being used in its legacy capacity
            return token

    def generate_token(self):
        now = datetime.datetime.now().strftime("%Y%m%d%H%M")
        public_auth_data = _combine(self._mailbox, self._nonce,
                                    self._nonce_count, now)
        private_auth_data = _combine(self._mailbox, self._nonce,
                                     self._nonce_count, self._password, now)
        myhash = hmac.HMAC(self._key, private_auth_data.encode("ASCII"),
                           sha256).hexdigest()
        self._nonce_count += 1
        return "NHSMESH {public_auth_data}:{myhash}".format(**locals())


# Preserve old name, even though it's part of the API now
_AuthTokenGenerator = AuthTokenGenerator


def _combine(*elements):
    return ":".join(str(x) for x in elements)



from __future__ import division
import io
import os
import six
import warnings
import zlib


class IteratorMixin(object):
    """
    Produce a series of lines, from a readable that only exposes a `read` method
    """
    __block_size = 65536
    __line_iterator = None

    def __iter__(self):
        # We use a BytesIO as our buffer, so we can piggyback on its readlines method
        buff = io.BytesIO()
        while True:
            block = self.read(self.__block_size)
            buff.write(block)
            last_block = len(block) == 0
            buff.seek(0)
            lines = buff.readlines()
            for line in lines[:-1]:
                yield line
            if not last_block:
                buff.seek(0)
                buff.write(lines[-1])
                buff.truncate()
            else:
                yield lines[-1]
                break

    def readline(self):
        if not self.__line_iterator:
            self.__line_iterator = iter(self)
        try:
            return six.next(self.__line_iterator)
        except StopIteration:
            return b''

    def readlines(self):
        return list(iter(self))


class CloseUnderlyingMixin(object):
    def close(self):
        try:
            if hasattr(self._underlying, "close"):
                self._underlying.close()
        finally:
            self._underlying = None
            if hasattr(super(CloseUnderlyingMixin, self), "close"):
                super(CloseUnderlyingMixin, self).close()

    def __del__(self):
        self.close()
        if hasattr(super(CloseUnderlyingMixin, self), "__del__"):
            super(CloseUnderlyingMixin, self).__del__()

    def __enter__(self):
        if hasattr(super(CloseUnderlyingMixin, self), "__enter__"):
            return super(CloseUnderlyingMixin, self).__enter__()
        else:
            return self

    def __exit__(self, typ, value, traceback):
        self.close()
        if hasattr(super(CloseUnderlyingMixin, self), "__exit__"):
            return super(CloseUnderlyingMixin, self).__exit__(typ, value,
                                                              traceback)


class AbstractGzipStream(IteratorMixin, CloseUnderlyingMixin):
    """
    Wrap an existing readable, in a readable that produces a gzipped
    version of the underlying stream.
    """

    def __init__(self, underlying, block_size=65536):
        self._underlying = underlying
        self._buffer = io.BytesIO()
        self._block_size = block_size

    def _process_block(self, block):
        raise NotImplementedError()

    def _finish(self):
        raise NotImplementedError()

    def read(self, n=-1):
        if n == -1:
            n = None
        while True:
            # If underlying stream finished, just read from buffer
            if self._underlying is None:
                return self._buffer.read(n)

            # If underlying stream not finished, buffer is positioned at end,
            # ready for writing
            limit = self._buffer.tell()
            if n is not None and n != -1 and limit >= n:
                self._buffer.seek(0)
                try:
                    return self._buffer.read(n)
                finally:
                    remainder = self._buffer.read(limit - n)
                    self._buffer = io.BytesIO(remainder)
                    self._buffer.seek(0, 2)
            else:
                next_block = self._underlying.read(self._block_size)
                if len(next_block) > 0:
                    self._buffer.write(self._process_block(next_block))
                else:
                    self._buffer.write(self._finish())
                    self._buffer.seek(0)
                    self.close()

    def read_all(self):
        self.read()


class GzipCompressStream(AbstractGzipStream):
    """
    Wrap an existing readable, in a readable that produces a gzipped
    version of the underlying stream.
    """

    def __init__(self, underlying, block_size=65536):
        AbstractGzipStream.__init__(self, underlying, block_size)
        self._compress_obj = zlib.compressobj(
            9,  # level
            zlib.DEFLATED,  # method
            31  # wbits - gzip header, maximum window
        )

    def _process_block(self, block):
        return self._compress_obj.compress(block)

    def _finish(self):
        return self._compress_obj.flush(zlib.Z_FINISH)


class GzipDecompressStream(AbstractGzipStream):
    """
    Wrap an existing readable, in a readable that decompresses
    the underlying stream.
    """

    def __init__(self, underlying, block_size=65536):
        AbstractGzipStream.__init__(self, underlying, block_size)
        self._decompress_obj = zlib.decompressobj(
            47  # wbits - detect header, maximum window
        )

    def _process_block(self, block):
        return self._decompress_obj.decompress(block)

    def _finish(self):
        return self._decompress_obj.flush()


class SplitStream(CloseUnderlyingMixin):
    def __init__(self, data, chunk_size=75 * 1024 * 1024):
        if isinstance(data, bytes):
            self._underlying = io.BytesIO(data)
            self._length = len(data)
        elif hasattr(data, "info"):
            self._underlying = data
            self._length = int(data.info()["Content-Length"])
        elif hasattr(data, "fileno"):
            self._underlying = data
            self._length = os.fstat(data.fileno()).st_size
        elif hasattr(data, "_content_length"):
            self._underlying = data
            self._length = data._content_length
        elif isinstance(data, dict) and "Body" in data and "ContentLength" in data:
            self._underlying = data["Body"]
            self._length = data["ContentLength"]
        else:
            raise TypeError("data must be a bytes, file, or urllib response")
        self._chunk_size = chunk_size
        self._remaining = 0

    def __len__(self):
        return max(1,
                   (self._length + self._chunk_size - 1) // self._chunk_size)

    def __iter__(self):
        for i in range(len(self)):
            if self._remaining > 0:
                warnings.warn(
                    "moving to next chunk, despite unread content in buffer "
                    "- existing chunk will become invalid")
                self._underlying.read(self._remaining)
            self._remaining = min(self._chunk_size,
                                  self._length - i * self._chunk_size)
            yield _SplitChunk(self)


class _SplitChunk(IteratorMixin):
    def __init__(self, owner):
        self._owner = owner

    def read(self, n=-1):
        if n == -1:
            n = None
        if n is None or n > self._owner._remaining:
            n = self._owner._remaining
        try:
            return self._owner._underlying.read(n)
        finally:
            self._owner._remaining -= n

    def __len__(self):
        return self._owner._remaining


class CombineStreams(IteratorMixin):
    def __init__(self, streams):
        self._streams = iter(streams)
        self._current_stream = six.next(self._streams)

    def read(self, n=-1):
        if n == -1:
            n = None
        result = io.BytesIO()
        try:
            while True:
                data_read = self._current_stream.read(n)
                result.write(data_read)
                if n is None or len(data_read) < n:
                    self._close_current_stream()
                    self._current_stream = six.next(self._streams)
                    if n is not None:
                        n -= len(data_read)
                else:
                    return result.getvalue()
        except StopIteration:
            self._current_stream = io.BytesIO()  # Empty stream
            return result.getvalue()

    def close(self):
        self._close_current_stream()

    def _close_current_stream(self):
        try:
            self._current_stream.close()
        except:
            pass


class FiniteLengthStream(IteratorMixin, CloseUnderlyingMixin):
    def __init__(self, stream, length):
        self._underlying = stream
        self._remaining = length

    def read(self, n=-1):
        if n == -1 or n is None:
            n = self._remaining
        n = min(n, self._remaining)
        try:
            return self._underlying.read(n)
        finally:
            self._remaining -= n


class ChunkedStream(IteratorMixin, CloseUnderlyingMixin):
    def __init__(self, underlying):
        self._underlying = underlying
        self._buffer = io.BytesIO()

    def read(self, n=-1):
        if n == -1:
            n = None
        while True:
            # If underlying stream finished, just read from buffer
            if self._underlying is None:
                return self._buffer.read(n)

            # If underlying stream not finished, buffer is positioned at end,
            # ready for writing
            limit = self._buffer.tell()
            if n is not None and limit >= n:
                self._buffer.seek(0)
                try:
                    return self._buffer.read(n)
                finally:
                    remainder = self._buffer.read(limit - n)
                    self._buffer = io.BytesIO(remainder)
                    self._buffer.seek(0, 2)
            else:
                block_header = self._underlying.read(1)
                while not block_header.endswith(b"\r\n"):
                    block_header += self._underlying.read(1)
                block_length = int(block_header, 16)
                if block_length > 0:
                    self._buffer.write(self._underlying.read(block_length))
                    assert self._underlying.read(2) == b"\r\n"
                else:
                    self._buffer.seek(0)
                    self._underlying.close()
                    self._underlying = None


def stream_from_wsgi_environ(environ):
    if environ.get("CONTENT_LENGTH"):
        return FiniteLengthStream(environ["wsgi.input"],
                                  int(environ["CONTENT_LENGTH"]))
    elif environ.get("HTTP_TRANSFER_ENCODING") == "chunked":
        return ChunkedStream(environ["wsgi.input"])
    else:
        # terminated by close
        return environ["wsgi.input"]



from __future__ import absolute_import, print_function

import io
from unittest import TestCase, main
import mock
import requests
import signal
import traceback
import sys

from collections import namedtuple
from mesh_client import MeshClient, MeshError, default_ssl_opts, CombineStreams, LOCAL_MOCK_ENDPOINT
from mesh_client.mock_server import MockMeshApplication, MockMeshChunkRetryApplication, SlowMockMeshApplication
from six.moves.urllib.error import HTTPError

alice_mailbox = 'alice'
alice_password = 'password'
bob_mailbox = 'bob'
bob_password = 'password'


class MockResponse:
    def __init__(self, json_data, status_code):
        self.json_data = json_data
        self.status_code = status_code

    def json(self):
        return self.json_data


def print_stack_frames(signum=None, frame=None):
    for frame in sys._current_frames().values():
        traceback.print_stack(frame)
        print()


signal.signal(signal.SIGUSR1, print_stack_frames)


class TestError(Exception):
    pass


class MeshClientTest(TestCase):
    def run(self, result=None):
        try:
            with MockMeshApplication() as mock_app:
                self.mock_app = mock_app
                self.uri = mock_app.uri
                with MeshClient(
                        self.uri,
                        alice_mailbox,
                        alice_password,
                        max_chunk_size=5,
                        **default_ssl_opts) as alice:
                    self.alice = alice
                    self.bob = MeshClient(
                        self.uri,
                        bob_mailbox,
                        bob_password,
                        max_chunk_size=5,
                        **default_ssl_opts)
                    super(MeshClientTest, self).run(result)
        except HTTPError as e:
            print(e.read())
            print_stack_frames()
            print("Message store", self.mock_app.messages)
            raise
        except:
            print_stack_frames()
            print("Message store", self.mock_app.messages)
            raise

    def test_handshake(self):
        alice = self.alice

        hand_shook = alice.handshake()
        self.assertEqual(hand_shook, b"hello")

    def test_send_receive(self):
        alice = self.alice
        bob = self.bob

        message_id = alice.send_message(bob_mailbox, b"Hello Bob 1")
        self.assertEqual([message_id], bob.list_messages())
        self.assertEqual(1, bob.count_messages())
        msg = bob.retrieve_message(message_id)
        self.assertEqual(msg.read(), b"Hello Bob 1")
        self.assertEqual(msg.sender, "alice")
        self.assertEqual(msg.recipient, "bob")
        msg.acknowledge()
        self.assertEqual([], bob.list_messages())

    def test_send_receive_combine_streams_part1_multiple_of_chunk_size(self):
        alice = self.alice
        bob = self.bob

        part1_length = 10
        part2_length = 23

        stream = {
            'Body': CombineStreams([io.BytesIO(b"H"*part1_length), io.BytesIO((b"W"*part2_length))]),
            'ContentLength': part1_length + part2_length,
        }

        message_id = alice.send_message(bob_mailbox, stream)
        self.assertEqual([message_id], bob.list_messages())
        self.assertEqual(1, bob.count_messages())
        msg = bob.retrieve_message(message_id)
        self.assertEqual(msg.read(), b"H" * part1_length + b"W"*part2_length)
        self.assertEqual(msg.sender, "alice")
        self.assertEqual(msg.recipient, "bob")
        msg.acknowledge()
        self.assertEqual([], bob.list_messages())

    def test_send_receive_combine_streams_part1_not_multiple_of_chunk_size(self):
        alice = self.alice
        bob = self.bob

        part1_length = 4
        part2_length = 20

        stream = {
            'Body': CombineStreams([io.BytesIO(b"H"*part1_length), io.BytesIO((b"W"*part2_length))]),
            'ContentLength': part1_length + part2_length,
        }

        message_id = alice.send_message(bob_mailbox, stream)
        self.assertEqual([message_id], bob.list_messages())
        self.assertEqual(1, bob.count_messages())
        msg = bob.retrieve_message(message_id)
        self.assertEqual(msg.read(), b"H" * part1_length + b"W"*part2_length)
        self.assertEqual(msg.sender, "alice")
        self.assertEqual(msg.recipient, "bob")
        msg.acknowledge()
        self.assertEqual([], bob.list_messages())

    def test_line_by_line(self):
        alice = self.alice
        bob = self.bob

        message_id = alice.send_message(bob_mailbox, b"Hello Bob 1\nHello Bob 2")
        self.assertEqual([message_id], bob.list_messages())
        msg = bob.retrieve_message(message_id)
        self.assertEqual(list(iter(msg)), [b"Hello Bob 1\n", b"Hello Bob 2"])

    def test_readline(self):
        alice = self.alice
        bob = self.bob

        message_id = alice.send_message(bob_mailbox, b"Hello Bob 1\nHello Bob 2")
        self.assertEqual([message_id], bob.list_messages())
        msg = bob.retrieve_message(message_id)
        self.assertEqual(msg.readline(), b"Hello Bob 1\n")

    def test_readlines(self):
        alice = self.alice
        bob = self.bob

        message_id = alice.send_message(bob_mailbox, b"Hello Bob 1\nHello Bob 2")
        self.assertEqual([message_id], bob.list_messages())
        msg = bob.retrieve_message(message_id)
        self.assertEqual(msg.readlines(), [b"Hello Bob 1\n", b"Hello Bob 2"])

    def test_transparent_compression(self):
        alice = self.alice
        bob = self.bob

        print("Sending")
        alice._transparent_compress = True
        message_id = alice.send_message(
            bob_mailbox, b"Hello Bob Compressed")
        self.assertEqual([message_id], bob.list_messages())
        print("Receiving")
        msg = bob.retrieve_message(message_id)
        self.assertEqual(msg.read(), b"Hello Bob Compressed")
        self.assertEqual(msg.mex_header('from'), 'alice')
        msg.acknowledge()
        self.assertEqual([], bob.list_messages())

    def test_iterate_and_context_manager(self):
        alice = self.alice
        bob = self.bob

        alice.send_message(bob_mailbox, b"Hello Bob 2")
        alice.send_message(bob_mailbox, b"Hello Bob 3")
        messages_read = 0
        for (msg, expected) in zip(bob.iterate_all_messages(),
                                   [b"Hello Bob 2", b"Hello Bob 3"]):
            with msg:
                self.assertEqual(msg.read(), expected)
                messages_read += 1
        self.assertEqual(2, messages_read)
        self.assertEqual([], bob.list_messages())

    def test_context_manager_failure(self):
        alice = self.alice
        bob = self.bob

        message_id = alice.send_message(bob_mailbox, b"Hello Bob 4")
        try:
            with bob.retrieve_message(message_id) as msg:
                self.assertEqual(msg.read(), b"Hello Bob 4")
                raise TestError()
        except TestError:
            pass
        self.assertEqual([message_id], bob.list_messages())

    def test_optional_args(self):
        alice = self.alice
        bob = self.bob

        message_id = alice.send_message(
            bob_mailbox,
            b"Hello Bob 5",
            subject="Hello World",
            filename="upload.txt",
            local_id="12345",
            message_type="DATA",
            process_id="321",
            workflow_id="111",
            encrypted=False,
            compressed=False)

        with bob.retrieve_message(message_id) as msg:
            self.assertEqual(msg.subject, "Hello World")
            self.assertEqual(msg.filename, "upload.txt")
            self.assertEqual(msg.local_id, "12345")
            self.assertEqual(msg.message_type, "DATA")
            self.assertEqual(msg.process_id, "321")
            self.assertEqual(msg.workflow_id, "111")
            self.assertFalse(msg.encrypted)
            self.assertFalse(msg.compressed)

        message_id = alice.send_message(
            bob_mailbox, b"Hello Bob 5", encrypted=True, compressed=True)

        with bob.retrieve_message(message_id) as msg:
            self.assertTrue(msg.encrypted)
            self.assertTrue(msg.compressed)

    def test_tracking(self):
        alice = self.alice
        bob = self.bob
        tracking_id = 'Message1'
        msg_id = alice.send_message(bob_mailbox, b'Hello World', local_id=tracking_id)
        self.assertEqual(alice.get_tracking_info(tracking_id)['status'], 'Accepted')
        bob.acknowledge_message(msg_id)
        self.assertEqual(alice.get_tracking_info(tracking_id)['status'], 'Acknowledged')

    def test_endpoint_lookup(self):
        result = self.alice.lookup_endpoint('ORG1', 'WF1')
        result_list = result['results']
        self.assertEqual(len(result_list), 1)
        self.assertEqual(result_list[0]['address'], 'ORG1HC001')
        self.assertEqual(result_list[0]['description'], 'ORG1 WF1 endpoint')
        self.assertEqual(result_list[0]['endpoint_type'], 'MESH')


    def test_error_handling(self):
        alice = self.alice
        with self.assertRaises(MeshError):
            alice.send_message(bob_mailbox, b"")


class EndpointTest(TestCase):
    def test_handshake(self):
        with MockMeshApplication() as mock_app:
            endpoint = LOCAL_MOCK_ENDPOINT._replace(url=mock_app.uri)
            client = MeshClient(
                endpoint,
                alice_mailbox,
                alice_password,
                max_chunk_size=5)

            hand_shook = client.handshake()
            self.assertEqual(hand_shook, b"hello")


class TimeoutTest(TestCase):
    def test_timeout(self):
        with SlowMockMeshApplication() as mock_app:
            endpoint = LOCAL_MOCK_ENDPOINT._replace(url=mock_app.uri)
            client = MeshClient(
                endpoint,
                alice_mailbox,
                alice_password,
                max_chunk_size=5,
                timeout=0.5)
            with self.assertRaises(requests.exceptions.Timeout):
                client.list_messages()



class MeshChunkRetryClientTest(TestCase):
    def run(self, result=None):
        try:
            with  MockMeshChunkRetryApplication() as mock_app:
                self.mock_app = mock_app
                self.uri = mock_app.uri
                self.alice = MeshClient(
                    self.uri,
                    alice_mailbox,
                    alice_password,
                    max_chunk_size=5,
                    max_chunk_retries=3,
                    **default_ssl_opts)
                self.bob = MeshClient(
                    self.uri,
                    bob_mailbox,
                    bob_password,
                    max_chunk_size=5,
                    max_chunk_retries=3,
                    **default_ssl_opts)
                super(MeshChunkRetryClientTest, self).run(result)
        except HTTPError as e:
            print(e.read())
            print_stack_frames()
            print("Message store", self.mock_app.messages)
            raise
        except:
            print_stack_frames()
            print("Message store", self.mock_app.messages)
            raise

    def _count_chunk_retry_call_counts(self, mocked_post):
        counts = {}
        for call in mocked_post.call_args_list:
            chunk = int(call.kwargs['headers']['Mex-Chunk-Range'].split(':')[0])
            if chunk not in counts.keys():
                counts[chunk] = 0
            else:
                counts[chunk] += 1
        return counts


    def test_chunk_retries(self):
        alice = self.alice
        bob = self.bob
        mock_post = mock.Mock(wraps=alice._session.post)
        alice._session.post = mock_post

        chunk_options = namedtuple('Chunk', 'chunk_num num_retry_attempts')
        options = [chunk_options(2, 2)]
        self.mock_app.set_chunk_retry_options(options)

        message_id = alice.send_message(bob_mailbox, b"Hello World")

        received = bob.retrieve_message(message_id).read()
        self.assertEqual(received, b'Hello World')

        chunk_retry_call_counts = self._count_chunk_retry_call_counts(mock_post)
        self.assertEqual(chunk_retry_call_counts[1], 0)
        self.assertEqual(chunk_retry_call_counts[2], 3)
        self.assertEqual(chunk_retry_call_counts[3], 0)

    def test_chunk_all_retries_fail(self):
        alice = self.alice
        mock_post = mock.Mock(wraps=alice._session.post)
        alice._session.post = mock_post

        chunk_options = namedtuple('Chunk', 'chunk_num num_retry_attempts')
        options = [chunk_options(2, 3)]
        self.mock_app.set_chunk_retry_options(options)

        self.assertRaises(requests.exceptions.HTTPError, alice.send_message, bob_mailbox, b"Hello World")

        chunk_retry_call_counts = self._count_chunk_retry_call_counts(mock_post)
        self.assertEqual(chunk_retry_call_counts[1], 0)
        self.assertEqual(chunk_retry_call_counts[2], 3)

    def test_chunk_retries_with_file(self):
        alice = self.alice
        bob = self.bob
        mock_post = mock.Mock(wraps=alice._session.post)
        alice._session.post = mock_post

        chunk_options = namedtuple('Chunk', 'chunk_num num_retry_attempts')
        options = [chunk_options(2, 2)]
        self.mock_app.set_chunk_retry_options(options)

        message_id = alice.send_message(bob_mailbox, open("test_chunk_retry_file", 'rb'))

        received = bob.retrieve_message(message_id).read()
        self.assertEqual(received, b'test1 test2 test3')

        chunk_retry_call_counts = self._count_chunk_retry_call_counts(mock_post)
        self.assertEqual(chunk_retry_call_counts[1], 0)
        self.assertEqual(chunk_retry_call_counts[2], 3)
        self.assertEqual(chunk_retry_call_counts[3], 0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
from setuptools import setup
from os.path import dirname, join

with open(join(dirname(__file__), 'README.md')) as f:
    long_description = f.read()

setup(
    name='Mesh Client',
    description='Client for NHS Digital \'s MESH messaging system',
    url='https://github.com/jamespic/mesh-client',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='James Pickering',
    author_email='james.pickering@airelogic.com',
    packages=['mesh_client'],
    package_data={'mesh_client': ['*.pem']},
    install_requires=[
        'requests (>=2.9.0)',
        'six (>=1.10.0)'
    ],
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    entry_points={
        'console_scripts': [
            'mesh_auth=mesh_client.mesh_auth:main',
            'mock_mesh_server=mesh_client.mock_server:main'
        ]
    },
    license='MIT',
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*',
    classifiers=[
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ]
)
